from typing import Tuple, List
import glob
from functools import reduce

from blue_objects import objects

from blue_geo.watch.targets import Target
from blue_geo.catalog.generic.generic.scope import raster_suffix
from blue_geo.logger import logger


def load_watch(object_name: str) -> Tuple[bool, Target, List[str]]:
    success, target = Target.load(object_name)

    list_of_files = sorted(
        reduce(
            lambda x, y: x + y,
            [
                glob.glob(
                    objects.path_of(
                        f"*{suffix}",
                        object_name,
                    )
                )
                for suffix in raster_suffix
            ],
        )
    )
    logger.info("{} file(s) to process.".format(len(list_of_files)))

    return success, target, list_of_files



from typing import List, Dict


class GenericCatalog:
    name: str = "generic"

    url: Dict[str, str] = {
        "keyword": "url",
    }

    def __init__(self):
        pass

    def get_list_of_collections(self) -> List[str]:
        return []

    def urls_as_str(self) -> List[str]:
        return sorted(
            [
                " - [{}]({})".format(
                    title.replace("-", " "),
                    url,
                )
                for title, url in self.url.items()
            ]
        )


class VoidCatalog(GenericCatalog):
    name = "void"



from typing import Tuple, List, Dict, Any
import matplotlib.pyplot as plt
from datetime import datetime
from collections import Counter
import geopandas as gpd
from geojson import Point

from blueness import module
from blue_options import string
from blue_objects import file, objects
from blue_objects.metadata import post_to_object

from blue_geo import NAME, env, fullname
from blue_geo.catalog.generic import GenericDatacube
from blue_geo.catalog.ukraine_timemap.classes import UkraineTimemapCatalog
from blue_geo.catalog.generic.generic.scope import DatacubeScope
from blue_geo.host import signature
from blue_geo.logger import logger

API_URL = "https://bellingcat-embeds.ams3.cdn.digitaloceanspaces.com/production/ukr/timemap/api.json"

DESCRIPTION = "Civilian Harm in Ukraine TimeMap"


NAME = module.name(__file__, NAME)


class UkraineTimemapDatacube(GenericDatacube):
    name = "ukraine_timemap"

    catalog = UkraineTimemapCatalog()

    QGIS_template = env.BLUE_GEO_QGIS_TEMPLATE_UKRAINE_TIMEMAP

    def ingest(
        self,
        dryrun: bool = False,
        overwrite: bool = False,
        do_save: bool = True,
        scope: str = "metadata",
        do_visualize: bool = True,
        log: bool = True,
    ) -> Tuple[bool, gpd.GeoDataFrame]:
        success, _ = super().ingest(dryrun, overwrite, scope)
        if not success:
            return success, gpd.GeoDataFrame()

        filename = objects.path_of(
            "ukraine_timemap.json",
            self.datacube_id,
            create=True,
        )

        gdf = gpd.GeoDataFrame()
        metadata: Dict[str, Any] = {
            "description": DESCRIPTION,
            "created_by": "-".join(signature()),
            "creation_date": string.pretty_date(),
        }

        success = file.download(API_URL, filename)
        if not success:
            return success, gdf

        success, list_of_events = file.load_json(filename)
        if not success:
            return success, gdf
        logger.info("{:,} event(s) ingested from the api.".format(len(list_of_events)))
        metadata["api_count"] = len(list_of_events)

        records = []
        failure_count = 0
        for event in list_of_events:
            try:
                point = Point(
                    (
                        float(event["longitude"]),
                        float(event["latitude"]),
                    )
                )
                record = {"geometry": point}

                for keyword, value in event.items():
                    record[keyword] = (
                        ", ".join(value) if isinstance(value, list) else value
                    )

                record["date_obj"] = datetime.strptime(event["date"], "%Y-%m-%d").date()

            except Exception as e:
                logger.warning(f"ingest failed: {e}: {event}")
                failure_count += 1
                continue

            records.append(record)

        if not records:
            logger.error("no records to ingest.")
            return False, gdf

        gdf = gpd.GeoDataFrame(records)

        gdf.set_crs(epsg=4326, inplace=True)  # WGS 84

        gdf = gdf.sort_values(by="date_obj", ascending=False)

        logger.info("{:,} event(s) -ingested-> gdf.".format(len(gdf)))
        metadata["ingested_count"] = len(gdf)
        if failure_count:
            logger.warning(f"{failure_count:,} event(s) failed to ingest.")
        metadata["failure_count"] = failure_count

        histogram = Counter(list(gdf["date_obj"].values))

        dates = sorted(histogram.keys())
        logger.info(
            "{:,} day(s) of events, starting {}, until {}.".format(
                len(dates),
                min(dates),
                max(dates),
            )
        )
        metadata["range"] = [min(dates), max(dates)]

        if do_visualize:
            values = [histogram[date] for date in dates]

            plt.figure(figsize=(10, 5))
            plt.bar(dates, values, color="blue")
            plt.xlabel(
                " | ".join(
                    [
                        "Date",
                        self.datacube_id,
                        fullname(),
                    ]
                )
            )
            plt.ylabel("# Events / Day")
            plt.title(DESCRIPTION)

            date_count = 20
            if len(dates) > date_count:
                selected_dates = [
                    dates[i] for i in range(0, len(dates), len(dates) // date_count)
                ]
                if dates[-1] not in selected_dates:
                    selected_dates.append(dates[-1])
            else:
                selected_dates = dates
            plt.xticks(selected_dates, rotation=45)

            plt.tight_layout()
            plt.grid(True)

            if do_save:
                file.save_fig(
                    objects.path_of("ukraine_timemap.png", self.datacube_id),
                    log=log,
                )

        gdf["date"] = gdf["date_obj"].apply(
            lambda d: f"{d.year}/{d.month:02}/{d.day:02}"
        )

        gdf = gdf.drop(columns=["date_obj"])

        if do_save and not gdf.empty:
            if not file.save_geojson(
                objects.path_of("ukraine_timemap.geojson", self.datacube_id),
                gdf,
                log=log,
            ) or not file.save_yaml(
                objects.path_of("metadata.yaml", self.datacube_id),
                metadata,
                log=log,
            ):
                return False, gdf

        return True, gdf

    def list_of_files(
        self,
        scope: DatacubeScope = DatacubeScope("all"),
        verbose: bool = False,
    ) -> List[str]:
        return [
            f"ukraine_timemap.{extension}"
            for extension in [
                "geojson",
                "json",
                "png",
            ]
        ]

    @classmethod
    def query(cls, object_name: str) -> bool:
        logger.info(f"🔎 {cls.__name__}.query -> {object_name}")

        datacube = cls(
            "datacube-ukraine_timemap-{}".format(
                string.pretty_date(as_filename=True),
            )
        )

        logger.info(f"🧊 {datacube.description}")

        return post_to_object(
            object_name,
            "datacube_id",
            [datacube.datacube_id],
        )



from typing import List
import argparse

from blueness import module
from blueness.argparse.generic import sys_exit

from blue_geo import NAME
from blue_geo.watch.targets.classes import TargetList, Target
from blue_geo.help.watch.targets import get_what_list
from blue_geo.logger import logger

NAME = module.name(__file__, NAME)

list_of_tasks = "get|list|save|test"


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help=list_of_tasks,
)
parser.add_argument(
    "--what",
    default="",
    type=str,
    help=get_what_list,
)
parser.add_argument(
    "--count",
    type=int,
    default=-1,
    help="-1: all",
)
parser.add_argument(
    "--log",
    default=1,
    type=int,
    help="0|1",
)
parser.add_argument(
    "--including_versions",
    default=1,
    type=int,
    help="0|1",
)
parser.add_argument(
    "--delim",
    type=str,
    default=",",
)
parser.add_argument(
    "--target_name",
    type=str,
)
parser.add_argument(
    "--object_name",
    type=str,
)
parser.add_argument(
    "--catalog_name",
    type=str,
)
parser.add_argument(
    "--collection",
    type=str,
)
args = parser.parse_args()

delim = " " if args.delim == "space" else args.delim

target_list = TargetList()

target = target_list.get(
    args.target_name,
    including_versions=args.including_versions == 1,
)

success = args.task in list_of_tasks
if args.task == "get":
    output: str = ""

    if args.what == "catalog":
        output = target.catalog
    elif args.what == "collection":
        output = target.collection
    elif args.what == "exists":
        output = "1" if target.name else "0"
    elif args.what == "one_liner":
        output = target.one_liner
    elif args.what == "query_args":
        output = target.query_args_as_str()
    else:
        success = False

    if success:
        if args.log:
            logger.info(output)
        else:
            print(output)
elif args.task == "list":
    output = target_list.get_list(
        catalog_name=args.catalog_name,
        collection=args.collection,
        including_versions=args.including_versions == 1,
    )

    if args.count != -1:
        output = output[: args.count]

    if success:
        if args.log:
            logger.info("{:,} target(s).".format(len(output)))
            for index, target_name in enumerate(output):
                logger.info(f"#{index: 4} - {target_name}")
        else:
            print(delim.join(output))
elif args.task == "save":
    success = target.save(args.object_name)
elif args.task == "test":
    success = target_list.test()
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import Dict, Tuple, List
import copy
import geopandas as gpd
from shapely.geometry import Polygon
from functools import reduce
from tqdm import tqdm

from blue_objects import file, objects

from blue_geo.env import BLUE_GEO_WATCH_TARGET_LIST
from blue_geo.logger import logger


class Target:
    def __init__(
        self,
        name: str = "",
        catalog: str = "",
        collection: str = "",
        params: Dict[str, str] = {},
        query_args: Dict[str, str] = {},
        urls: Dict[str, str] = {},
        versions: Dict[str, str] = {},
    ) -> None:
        self.name: str = name

        self.catalog = catalog
        self.collection = collection

        self.params = copy.deepcopy(params)
        self.query_args = copy.deepcopy(query_args)

        self.urls = copy.deepcopy(urls)

        self.versions = copy.deepcopy(versions)

    @classmethod
    def from_dict(
        cls,
        name: str,
        data: Dict,
    ) -> "Target":
        return cls(
            name=name,
            catalog=data.get("catalog", ""),
            collection=data.get("collection", ""),
            params=copy.deepcopy(data.get("params", {})),
            query_args=copy.deepcopy(data.get("query_args", {})),
            urls=copy.deepcopy(data.get("urls", {})),
            versions=copy.deepcopy(data.get("versions", {})),
        )

    def get_version(
        self,
        version: str,
        is_full_name: bool = False,
    ) -> "Target":
        if is_full_name:
            version = version[len(self.name) + 1 :]

        if version not in self.versions:
            return Target()

        version_updates = self.versions.get(version, {})

        params = copy.deepcopy(self.params)
        params.update(version_updates.get("params", {}))

        query_args = copy.deepcopy(self.query_args)
        query_args.update(version_updates.get("query_args", {}))

        return Target(
            name=f"{self.name}-{version}",
            catalog=self.catalog,
            collection=self.collection,
            params=params,
            query_args=query_args,
            urls=self.urls,
        )

    @property
    def lat_and_lon(self) -> str:
        lat = self.query_args["lat"]
        lon = self.query_args["lon"]

        return '`lat: {:.04f}"{}`, `lon: {:.04f}"{}`.'.format(
            abs(lat),
            "N" if lat > 0 else "S",
            abs(lon),
            "E" if lon > 0 else "W",
        )

    @classmethod
    def load(cls, object_name: str) -> Tuple[bool, "Target"]:
        success, metadata = file.load_yaml(
            objects.path_of(
                "target/metadata.yaml",
                object_name,
            )
        )

        return success, Target.from_dict(
            metadata.get("name", ""),
            metadata,
        )

    @property
    def one_liner(self) -> str:
        return "{}: {} | {}/{} | {} | {}{}".format(
            self.__class__.__name__,
            self.name,
            self.catalog,
            self.collection,
            self.query_args_as_str(" | "),
            " | ".join([f"{param}={value}" for param, value in self.params.items()]),
            (
                " | {} version(s): {}".format(
                    len(self.versions),
                    ",".join(self.versions.keys()),
                )
                if self.versions
                else ""
            ),
        )

    def query_args_as_str(self, delim: str = " ") -> str:
        return delim.join(
            [f"--{argument} {value}" for argument, value in self.query_args.items()]
        )

    def save(self, object_name: str) -> bool:
        if not file.save_yaml(
            objects.path_of(
                "target/metadata.yaml",
                object_name,
                create=True,
            ),
            self.__dict__,
        ):
            return False

        lat = self.query_args.get("lat", 0)
        lon = self.query_args.get("lon", 0)
        width = self.params.get("width", 0.1)
        height = self.params.get("height", 0.1)

        half_width = width / 2
        half_height = height / 2

        top_left = (lon - half_width, lat + half_height)
        top_right = (lon + half_width, lat + half_height)
        bottom_right = (lon + half_width, lat - half_height)
        bottom_left = (lon - half_width, lat - half_height)

        polygon = Polygon([top_left, top_right, bottom_right, bottom_left, top_left])

        gdf = gpd.GeoDataFrame(
            {
                "id": ["1"],
                "description": [self.name],
                "geometry": [polygon],
            },
            crs="EPSG:4326",
        )

        return file.save_geojson(
            objects.path_of("target/shape.geojson", object_name),
            gdf,
            log=True,
        )

    def urls_as_str(self) -> List[str]:
        return sorted(
            [
                " - [{}]({}){}".format(
                    title,
                    url.split(",", 1)[0],
                    (
                        ": {}".format(
                            self.lat_and_lon
                            if title == "Google Maps"
                            else url.split(",", 1)[1].strip()
                        )
                        if "," in url or title == "Google Maps"
                        else ""
                    ),
                )
                for title, url in self.urls.items()
            ]
        )


class TargetList:
    object_name = BLUE_GEO_WATCH_TARGET_LIST

    def __init__(
        self,
        load: bool = True,
        download: bool = False,
    ) -> None:
        self.list_of_targets: Dict[str, Target] = {}

        if load:
            assert self.load(download)

    @classmethod
    def filename(cls) -> str:
        return objects.path_of(
            filename="metadata.yaml",
            object_name=cls.object_name,
        )

    def get(
        self,
        target_name: str,
        including_versions: bool = True,
    ) -> Target:
        if not target_name:
            return Target()

        if target_name in self.list_of_targets:
            return self.list_of_targets[target_name]

        if not including_versions:
            return Target()

        for existing_target in self.list_of_targets.values():
            if target_name.startswith(existing_target.name):
                return existing_target.get_version(
                    target_name,
                    is_full_name=True,
                )

        return Target()

    def get_list(
        self,
        catalog_name: str = "",
        collection: str = "",
        including_versions: bool = True,
    ) -> List[str]:
        return sorted(
            reduce(
                lambda x, y: x + y,
                [
                    [target.name]
                    + (
                        [f"{target.name}-{version}" for version in target.versions]
                        if including_versions
                        else []
                    )
                    for target in self.list_of_targets.values()
                    if (target.catalog == catalog_name or not catalog_name)
                    and (target.collection == collection or not collection)
                ],
                [],
            )
        )

    def load(
        self,
        download: bool = False,
    ) -> bool:
        if download:
            if not objects.download(
                object_name=self.object_name,
                filename="metadata.yaml",
            ):
                return False

        _, targets = file.load_yaml(self.filename(), ignore_error=True)

        self.list_of_targets = {
            target_name: Target.from_dict(
                target_name,
                target_data,
            )
            for target_name, target_data in targets.items()
        }

        return True

    def test(self) -> bool:
        list_of_targets = self.get_list(
            including_versions=True,
        )
        logger.info(f"testing {len(list_of_targets)} target(s).")

        for target_name in tqdm(list_of_targets):
            target = self.get(
                target_name,
                including_versions=True,
            )
            if not target.name:
                logger.error(f"bad target: {target_name}")
                return False

            try:
                logger.info(target.one_liner)
            except Exception as e:
                logger.error(f"bad target: {target_name} - {e}")
                return False

        return True



from typing import List

from notebooks_and_scripts import fullname as notebooks_ands_scripts_fullname
from abcli.host import signature as abcli_signature

from blue_geo import fullname


def signature() -> List[str]:
    return [
        fullname(),
        notebooks_ands_scripts_fullname(),
    ] + abcli_signature()



from typing import Tuple, Union
import urllib3
import requests
import json
from pystac_client import Client

from blueness import module

from blue_geo import NAME
from blue_geo.env import SKYFOX_ACCESS_TOKEN_URL, SKYFOX_CLIENT_ID, SKYFOX_CLIENT_SECRET
from blue_geo.catalog.generic.stac.classes import STACCatalog
from blue_geo.logger import logger

NAME = module.name(__file__, NAME)

# https://urllib3.readthedocs.io/en/1.26.x/advanced-usage.html#ssl-warnings
urllib3.disable_warnings()


class SkyFoxCatalog(STACCatalog):
    name = "SkyFox"

    url = {
        "account": "https://console.earthdaily.com/account",
        "api": "https://api.earthdaily.com/platform/v1/stac",
        "api-use-example": "https://github.com/earthdaily/EDA-Documentation/blob/gh-pages/API/APIUsage/earthplatform_stac_api_examples.py",
        "doc": "https://earthdaily.github.io/EDA-Documentation/",
        "platform": "https://console.earthdaily.com/platform",
        "signin": "https://console.earthdaily.com/mosaics/signin",
        "venus-bands": "https://un-regard-sur-la-terre.org/2017/08/les-premieres-images-du-satellite-ven-s-sont-arrivees.html",
        "venus-on-aws-open-data": "https://registry.opendata.aws/venus-l2a-cogs/",
        "venus-mission": "https://www.eoportal.org/satellite-missions/venus",
        "venus-wikipedia": "https://en.wikipedia.org/wiki/VEN%CE%BCS",
    }

    @classmethod
    def get_client(cls) -> Tuple[bool, Union[Client, None]]:
        success, token = cls.get_new_token()
        if not success:
            return False, None

        try:
            # https://earthdaily.github.io/EDA-Documentation/API/APIUsage/Python/#getting-the-authentication-token-for-pystac-client
            client = Client.open(
                cls.url["api"],
                headers={
                    "Authorization": f"Bearer {token}",
                },
            )
        except Exception as e:
            logger.error(e)
            return False, None

        return True, client

    @staticmethod
    # https://earthdaily.github.io/EDA-Documentation/API/APIUsage/Python/#getting-the-authentication-token-for-pystac-client
    def get_new_token() -> Tuple[bool, str]:
        token_req_payload = {"grant_type": "client_credentials"}
        response = requests.post(
            SKYFOX_ACCESS_TOKEN_URL,
            data=token_req_payload,
            verify=False,
            allow_redirects=False,
            auth=(SKYFOX_CLIENT_ID, SKYFOX_CLIENT_SECRET),
        )

        # https://chat.openai.com/c/6deb94d0-826a-48de-b5ef-f7d8da416c82
        # response.raise_for_status()
        if response.status_code // 100 != 2:
            logger.error(f"SkyFoxCatalog.get_new_token() failed: {response}")
            return False, ""

        tokens = json.loads(response.text)

        return True, tokens["access_token"]



from typing import List

from blue_options.terminal import show_usage


def help_version(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        "@gdal version",
        "show gdal version.",
        mono=mono,
    )


help_functions = {
    "version": help_version,
}
